// SPDX: BSD-2-Clause

use crate::utils::duration_to_string;
use anyhow;
use dirs;
use log::warn;
use serde::{self, Deserialize, Serialize};
use std::env;
use std::path::{Path, PathBuf};
use std::time::Duration;
use tokio::{fs, io::AsyncReadExt};

use crate::PROGRAM_NAME;

#[derive(Debug, Deserialize, Serialize)]
pub struct Config {
    pub zone_id: String,
    pub api_key: String,
    pub domains: Vec<Box<str>>,
    #[serde(default)]
    pub max_errors_in_row: Option<usize>,
    #[serde(with = "duration_format", default)]
    pub max_duration: Option<Duration>,
}

pub async fn get_config_path() -> Result<PathBuf, Vec<PathBuf>> {
    let mut tried_paths = Vec::with_capacity(2);

    match env::current_dir() {
        Ok(current_dir) => {
            let cwd_config = current_dir.join(format!("{}.toml", PROGRAM_NAME));

            let is_valid_path = fs::metadata(&cwd_config)
                .await
                .map_or_else(|_| false, |meta| meta.is_file());

            if is_valid_path {
                return Ok(cwd_config);
            } else {
                tried_paths.push(cwd_config);
            }
        }
        Err(e) => {
            warn!("Failed to get current working directory: {:?}", e);
        }
    }

    match dirs::config_dir() {
        Some(config_dir) => {
            let config_file = config_dir.join(PROGRAM_NAME).join("config.toml");

            let is_valid_path = fs::metadata(&config_file)
                .await
                .map_or_else(|_| false, |meta| meta.is_file());
            if is_valid_path {
                return Ok(config_file);
            } else {
                tried_paths.push(config_file);
            }
        }
        None => {
            warn!("No configuration directory found.");
        }
    }

    Err(tried_paths)
}

pub async fn read_config<P: AsRef<Path>>(path: &P) -> anyhow::Result<Config> {
    let mut file = fs::File::open(path).await?;
    let mut buf = String::new();
    file.read_to_string(&mut buf).await?;
    Ok(toml::from_str(&buf)?)
}

mod duration_format {
    use super::*;
    use serde::{self, Deserialize, Deserializer, Serializer};
    use std::time::Duration;

    pub fn serialize<S>(duration: &Option<Duration>, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        match duration.as_ref() {
            Some(duration) => serializer.serialize_str(duration_to_string(duration).trim()),
            None => serializer.serialize_none(),
        }
    }

    pub fn deserialize<'de, D>(deserializer: D) -> Result<Option<Duration>, D::Error>
    where
        D: Deserializer<'de>,
    {
        Option::<Box<str>>::deserialize(deserializer)?.map_or(Ok(None), |s| {
            parse_duration(&s)
                .map(Some)
                .map_err(serde::de::Error::custom)
        })
    }

    fn parse_duration(s: &str) -> Result<Duration, String> {
        let mut total_duration = Duration::new(0, 0);
        let units = [("d", 86400), ("h", 3600), ("m", 60), ("s", 1)];
        let mut remainder = s;

        for &(unit, factor) in &units {
            if let Some(idx) = remainder.find(unit) {
                let (value, rest) = remainder.split_at(idx);
                let value: u64 = value.trim().parse().map_err(|_| "Invalid number")?;
                total_duration += Duration::from_secs(value * factor);
                remainder = &rest[unit.len()..];
            }
        }

        if let Some(idx) = remainder.find("ns") {
            let (value, rest) = remainder.split_at(idx);
            let value: u32 = value.trim().parse().map_err(|_| "Invalid number")?;
            total_duration += Duration::new(0, value);
            remainder = &rest["ns".len()..];
        }

        if !remainder.trim().is_empty() {
            return Err("Invalid duration format".to_string());
        }

        Ok(total_duration)
    }
}
