// SPDX: BSD-2-Clause

use anyhow::{Context, Result};
use log::{error, info};
use reqwest::Client;
use std::{
    fmt,
    net::{IpAddr, Ipv4Addr},
};

use super::cloudflare_responses::{CloudflareResponse, DnsRecord};
use crate::get_current_public_ipv4;

pub struct CloudflareClient<A, Z>
where
    A: fmt::Display,
    Z: fmt::Display,
{
    client: Client,
    domains: Vec<Box<str>>,
    current_ip: Ipv4Addr,
    api_key: A,
    zone_id: Z,
}

impl<Api, Zone> CloudflareClient<Api, Zone>
where
    Api: fmt::Display,
    Zone: fmt::Display,
{
    pub async fn new(api_key: Api, zone_id: Zone, domains: Vec<Box<str>>) -> Result<Self> {
        let force_ipv4 = IpAddr::from([0, 0, 0, 0]);
        let client = reqwest::ClientBuilder::new()
            .local_address(force_ipv4)
            .build()?;

        let current_ip = get_current_public_ipv4(&client).await?;

        let this = Self {
            client,
            domains,
            current_ip,
            api_key,
            zone_id,
        };
        this.update_dns_records(current_ip).await?;
        Ok(this)
    }

    pub async fn check(&mut self) -> Result<()> {
        let new_ip = get_current_public_ipv4(&self.client).await?;
        if new_ip == self.current_ip {
            info!("IP '{}' is already set", new_ip);
            return Ok(());
        }
        info!(
            "IP has changed from '{}' -> '{}'",
            &self.current_ip, &new_ip
        );
        self.update_dns_records(new_ip).await?;
        self.current_ip = new_ip;

        Ok(())
    }

    async fn update_dns_records(&self, new_ip: Ipv4Addr) -> Result<()> {
        for domain in &self.domains {
            let records = match self.get_dns_records(domain).await {
                Ok(r) => r,
                Err(e) => {
                    error!(
                        "Could not fetch DNS records for domain '{}': {:?}",
                        &domain, &e
                    );
                    continue;
                }
            };

            let new_ip_s = new_ip.to_string().into_boxed_str();
            for mut record in records.into_iter() {
                if record.content == new_ip_s {
                    info!(
                        "On {}, ip already updated to '{}'",
                        &record.name, &record.content
                    );
                    continue;
                }
                info!(
                    "On {}, updating {}: '{}' -> '{}'",
                    &domain, record.name, record.content, &new_ip_s
                );
                if let Err(e) = self.update_dns_record(&mut record, new_ip_s.clone()).await {
                    error!(
                        "On {}, failed to update {}: '{}' -> '{}': {:?}",
                        &domain, record.name, record.content, &new_ip_s, &e
                    );
                    return Err(e);
                }
            }
        }

        Ok(())
    }

    async fn get_dns_records<D: fmt::Display>(&self, domain: &D) -> Result<Vec<DnsRecord>> {
        let url = format!(
            "https://api.cloudflare.com/client/v4/zones/{}/dns_records?type=A&name={}",
            self.zone_id, domain
        );

        let response = self
            .client
            .get(&url)
            .header("Authorization", format!("Bearer {}", self.api_key))
            .header("Content-Type", "application/json")
            .send()
            .await?
            .error_for_status()?
            .text()
            .await?;

        let cloudflare_response: CloudflareResponse =
            serde_json::from_str(&response).context("Failed to deserialize Cloudflare response")?;

        if !cloudflare_response.success {
            error!(
                "Cloudflare API returned errors: {:?}",
                cloudflare_response.errors
            );
            return Err(anyhow::anyhow!("Cloudflare API returned errors"));
        }

        cloudflare_response
            .result
            .context("Key 'result' not found in response")
    }

    async fn update_dns_record(&self, record: &mut DnsRecord, ip_content: Box<str>) -> Result<()> {
        let url = format!(
            "https://api.cloudflare.com/client/v4/zones/{}/dns_records/{}",
            self.zone_id, record.id
        );

        record.content = ip_content;
        self.client
            .put(&url)
            .header("Authorization", format!("Bearer {}", self.api_key))
            .header("Content-Type", "application/json")
            .json(&record)
            .send()
            .await?
            .error_for_status()?;

        Ok(())
    }
}
