// SPDX: BSD-2-Clause

use futures::future::{self, Either};
use log::{error, info};
use scopeguard::defer;
use tokio::time;

use dynip_cloudflare::{
    utils, CloudflareClient, ExitListener, MessageHandler, NetworkChangeListener,
    MAX_ERORS_IN_ROW_DEFAULT,
};

#[tokio::main]
async fn main() {
    dynip_cloudflare::init_logger();
    defer! {
        log::logger().flush();
    }
    let exit_listener = ExitListener::new();

    let config = match utils::get_config().await {
        Some(aux) => aux,
        None => return,
    };

    let mut cloudflare =
        match CloudflareClient::new(config.api_key, config.zone_id, config.domains).await {
            Ok(cloudflare) => cloudflare,
            Err(e) => {
                error!("Failed to initialize cloudflare client: {:?}", &e);
                return;
            }
        };

    let mut network_change_listener = match NetworkChangeListener::new() {
        Some(aux) => {
            info!("Listening for IPv4 address changes and interface connect/disconnect events...");
            aux
        }
        None => {
            error!("Failed to initialize networkchangelistener");
            return;
        }
    };

    let mut message_handler = MessageHandler::new(
        &mut cloudflare,
        config.max_errors_in_row.unwrap_or(MAX_ERORS_IN_ROW_DEFAULT),
    );

    let mut interval = config.max_duration.map(|duration| {
        let mut interval = time::interval(duration);
        interval.set_missed_tick_behavior(time::MissedTickBehavior::Delay);
        interval
    });

    while !exit_listener.should_exit() {
        let tick_future = match interval.as_mut() {
            Some(interval) => Either::Left(interval.tick()),
            None => Either::Right(future::pending::<tokio::time::Instant>()),
        };

        tokio::select! {
            _ = exit_listener.notified() => break,
            _ = tick_future => {
                if let Some(duration) = config.max_duration.as_ref() {
                    let duration_string = utils::duration_to_string(duration);
                    let log_string = format!("{} has passed since last check, checking...", duration_string.trim());
                    message_handler.log_and_check(Some(&log_string), Option::<&&str>::None).await;
                }
            }
            message = network_change_listener.next_message() => {
                if let Some((message, _)) = message {
                    if let Some(interval) = interval.as_mut() {
                        interval.reset();
                    }
                    message_handler.handle_message(message).await;
                }
            }
        }
    }
}
