// SPDX: BSD-2-Clause

use std::fmt;
use log::{debug, error, info};
use netlink_packet_core::{NetlinkMessage, NetlinkPayload};
use netlink_packet_route::RouteNetlinkMessage as RtnlMessage;

use crate::CloudflareClient;

pub struct MessageHandler<'a, Api, Zone>
where
    Api: fmt::Display,
    Zone: fmt::Display,
{
    cloudflare: &'a mut CloudflareClient<Api, Zone>,
    errs_counter: usize,
    errs_max: usize,
}

impl<'a, Api, Zone> MessageHandler<'a, Api, Zone>
where
    Api: fmt::Display,
    Zone: fmt::Display,
{
    pub fn new(cloudflare: &'a mut CloudflareClient<Api, Zone>, errs_max: usize) -> Self {
        Self {
            cloudflare,
            errs_counter: 0,
            errs_max,
        }
    }

    pub async fn handle_message(&mut self, message: NetlinkMessage<RtnlMessage>) -> Option<()> {
        match message.payload {
            NetlinkPayload::InnerMessage(RtnlMessage::NewAddress(msg)) => {
                self.log_and_check(Some("New IPv4 address"), Some(&msg))
                    .await
            }
            NetlinkPayload::InnerMessage(RtnlMessage::DelAddress(msg)) => {
                self.log_info("Deleted IPv4 address", &msg).await
            }
            NetlinkPayload::InnerMessage(RtnlMessage::NewLink(link)) => {
                self.log_and_check(Some("New link (interface connected)"), Some(&link))
                    .await
            }
            NetlinkPayload::InnerMessage(RtnlMessage::DelLink(link)) => {
                self.log_info("Deleted link (interface disconnected)", &link)
                    .await
            }
            _ => {
                debug!("Unhandled message payload: {:?}", message.payload);
                Some(())
            }
        }
    }

    pub async fn log_and_check<D, M>(&mut self, log_msg: Option<&D>, msg: Option<&M>) -> Option<()>
    where
        D: fmt::Display + ?Sized,
        M: fmt::Debug,
    {
        if let Some(s) = log_msg {
            info!("{}", s);
        }
        if let Some(m) = msg {
            if let Some(lm) = log_msg {
                debug!("{}: {:?}", lm, m);
            } else {
                debug!("{:?}", m);
            }
        }
        if let Err(e) = self.cloudflare.check().await {
            self.errs_counter += 1;
            error!(
                "Failed to check cloudflare ({}/{}): {:?}",
                self.errs_counter, self.errs_max, &e
            );
            if self.errs_counter >= self.errs_max {
                return None;
            }
        }
        Some(())
    }

    async fn log_info<M: fmt::Debug>(&self, log_msg: &str, msg: &M) -> Option<()> {
        info!("{}", log_msg);
        debug!("{:?} message: {:?}", log_msg, msg);
        Some(())
    }
}
